package ip;

import ifc.ifcDn;
import ifc.ifcEthTyp;
import ifc.ifcNull;
import ifc.ifcUp;
import java.util.List;
import pack.packHolder;
import tab.tabLabel;
import tab.tabLabelNtry;
import util.counter;
import util.debugger;
import util.logger;
import util.state;

/**
 * multiprotocol label switching (rfc3032) packet
 *
 * @author matecsaba
 */
public class ipMpls implements ifcUp {

    /**
     * size of header
     */
    public final static int size = 4;

    /**
     * mpls unicast ethertype
     */
    public final static int typeU = 0x8847;

    /**
     * mpls multicast ethertype
     */
    public final static int typeM = 0x8848;

    /**
     * ipv4 explicit null
     */
    public final static int labelExp4 = 0;

    /**
     * router alert
     */
    public final static int labelAlert = 1;

    /**
     * ipv6 explicit null
     */
    public final static int labelExp6 = 2;

    /**
     * implicit null
     */
    public final static int labelImp = 3;

    /**
     * entropy label
     */
    public final static int labelEntropy = 7;

    /**
     * generic associated label
     */
    public final static int labelGal = 13;

    /**
     * operation and maintenance
     */
    public final static int labelOam = 14;

    /**
     * extension label
     */
    public final static int labelExt = 15;

    private ifcDn lower = new ifcNull();

    private counter cntr = new counter();

    private ipFwd fwd4;

    private ipFwd fwd6;

    private ifcEthTyp fwdE;

    /**
     * create mpls handler
     *
     * @param fwdr4 ip4 forwarder
     * @param fwdr6 ip6 forwarder
     * @param ether ethertype handler
     */
    public ipMpls(ipFwd fwdr4, ipFwd fwdr6, ifcEthTyp ether) {
        fwd4 = fwdr4;
        fwd6 = fwdr6;
        fwdE = ether;
    }

    /**
     * register to ethtyp
     */
    public void register2eth() {
        fwdE.addET(typeM, "mplsMulti", this);
        fwdE.updateET(typeM, this);
        fwdE.addET(typeU, "mplsUni", this);
        fwdE.updateET(typeU, this);
    }

    /**
     * unregister from ethtyp
     */
    public void unregister2eth() {
        fwdE.delET(typeM);
        fwdE.delET(typeU);
    }

    /**
     * send one packet
     *
     * @param pck packet to send
     */
    public void send2eth(packHolder pck) {
        lower.sendPack(pck);
    }

    /**
     * parse mpls header
     *
     * @param pck packet to parse
     * @return false if successful, true if error happened
     */
    public static boolean parseMPLSheader(packHolder pck) {
        if (pck.dataSize() < size) {
            return true;
        }
        pck.MPLSlabel = pck.msbGetD(0) >>> 12;
        pck.MPLSttl = pck.getByte(3);
        int i = pck.getByte(2);
        pck.MPLSbottom = (i & 1) != 0;
        pck.MPLSexp = (i >>> 1) & 7;
        pck.getSkip(size);
        if (debugger.ipMPLStraf) {
            logger.debug("rx label=" + pck.MPLSlabel + " exp=" + pck.MPLSexp + " ttl=" + pck.MPLSttl + " bottom=" + pck.MPLSbottom);
        }
        return false;
    }

    /**
     * create mpls header
     *
     * @param pck packet to update
     */
    public static void createMPLSheader(packHolder pck) {
        pck.merge2beg();
        if (debugger.ipMPLStraf) {
            logger.debug("tx label=" + pck.MPLSlabel + " exp=" + pck.MPLSexp + " ttl=" + pck.MPLSttl + " bottom=" + pck.MPLSbottom);
        }
        if (pck.MPLSlabel == labelImp) {
            return;
        }
        int i = pck.MPLSlabel << 12;
        i |= pck.MPLSttl & 0xff;
        i |= (pck.MPLSexp & 7) << 9;
        if (pck.MPLSbottom) {
            i |= 0x100;
        }
        pck.msbPutD(0, i);
        pck.putSkip(size);
        pck.merge2beg();
        pck.MPLSbottom = false;
    }

    /**
     * create mpls labels
     *
     * @param pck packet to update
     * @param labs labels to add
     */
    public static void createMPLSlabels(packHolder pck, List<Integer> labs) {
        for (int i = labs.size() - 1; i >= 0; i--) {
            pck.MPLSlabel = labs.get(i);
            ipMpls.createMPLSheader(pck);
        }
    }

    /**
     * create error report
     *
     * @param pck packet to report
     * @param lab label to report
     * @param res reason code
     * @return false if succeed, true if error
     */
    public static boolean createError(packHolder pck, tabLabelNtry lab, counter.reasons res) {
        if (!pck.MPLSbottom) {
            return true;
        }
        if (lab.forwarder == null) {
            return true;
        }
        ipFwdIface ifc = new ipFwdIface(lab.iface, null);
        ifc = lab.forwarder.ifaces.find(ifc);
        if (ifc == null) {
            return true;
        }
        if (!ifc.unreachables) {
            return true;
        }
        if (lab.forwarder.ipCore.parseIPheader(pck, true)) {
            return true;
        }
        if (debugger.ipFwdTraf) {
            logger.debug("drop " + pck.IPsrc + " -> " + pck.IPtrg + " pr=" + pck.IPprt + " reason=" + counter.reason2string(res));
        }
        if (lab.forwarder.icmpCore.createError(pck, res, ifc.addr.copyBytes())) {
            return true;
        }
        lab.forwarder.ipCore.createIPheader(pck);
        pck.INTupper = -1;
        ipMpls.beginMPLSfields(pck, lab.forwarder.mplsPropTtl);
        return false;
    }

    /**
     * start mpls header
     *
     * @param pck packet to update
     * @param prop propagate ip ttl
     */
    public static void beginMPLSfields(packHolder pck, boolean prop) {
        pck.MPLSbottom = true;
        pck.MPLSexp = pck.IPtos >>> 5;
        if (prop) {
            pck.MPLSttl = pck.IPttl;
        } else {
            pck.MPLSttl = 255;
        }
    }

    public void setParent(ifcDn parent) {
        lower = parent;
    }

    public counter getCounter() {
        return cntr;
    }

    public void setState(state.states stat) {
    }

    public void closeUp() {
    }

    public void recvPack(packHolder pck) {
        cntr.rx(pck);
        int ethTyp = pck.msbGetW(0); // ethertype
        switch (ethTyp) {
            case typeU:
            case typeM:
                break;
            default:
                cntr.drop(pck, counter.reasons.badEthTyp);
                return;
        }
        pck.getSkip(2);
        for (;;) {
            if (parseMPLSheader(pck)) {
                break;
            }
            pck.MPLSttl--;
            if (pck.MPLSttl < 0) {
                break;
            }
            switch (pck.MPLSlabel) {
                case labelImp:
                case labelExp4:
                    if (!pck.MPLSbottom) {
                        continue;
                    }
                    if (fwd4 == null) {
                        cntr.drop(pck, counter.reasons.noIface);
                        return;
                    }
                    fwd4.mplsRxPack(fwd4.commonLabel, pck);
                    return;
                case labelExp6:
                    if (!pck.MPLSbottom) {
                        continue;
                    }
                    if (fwd6 == null) {
                        cntr.drop(pck, counter.reasons.noIface);
                        return;
                    }
                    fwd6.mplsRxPack(fwd6.commonLabel, pck);
                    return;
                case labelEntropy:
                    if (parseMPLSheader(pck)) {
                        cntr.drop(pck, counter.reasons.badLen);
                        return;
                    }
                    if (!pck.MPLSbottom) {
                        continue;
                    }
                    if (fwd4 == null) {
                        cntr.drop(pck, counter.reasons.noIface);
                        return;
                    }
                    fwd4.mplsRxPack(fwd4.commonLabel, pck);
                    return;
                case labelAlert:
                    continue;
                default:
                    break;
            }
            tabLabelNtry ntry = tabLabel.find(pck.MPLSlabel);
            if (ntry == null) {
                logger.info("received invalid label (" + pck.MPLSlabel + ") on " + lower);
                cntr.drop(pck, counter.reasons.notInTab);
                return;
            }
            ntry.cntr.rx(pck);
            if (ntry.forwarder == null) {
                cntr.drop(pck, counter.reasons.noRoute);
                return;
            }
            if (ntry.nextHop != null) {
                ntry.forwarder.mplsRxPack(ntry, pck);
                return;
            }
            if (pck.MPLSbottom) {
                ntry.forwarder.mplsRxPack(ntry, pck);
                return;
            }
        }
    }

}
